/********************************************************************************************
 *
 * CHECK_NRPE.C - NRPE Plugin For Nagios
 * Copyright (c) 1999-2008 Ethan Galstad (nagios@nagios.org)
 * License: GPL
 *
 * Last Modified: 11-11-2011
 *
 * Command line: CHECK_NRPE -H <host_address> [-p port] [-c command] [-to to_sec]
 *
 * Description:
 *
 * This plugin will attempt to connect to the NRPE daemon on the specified server and port.
 * The daemon will attempt to run the command defined as [command].  Program output and
 * return code are sent back from the daemon and displayed as this plugin's own output and
 * return code.
 *
 ********************************************************************************************/

#include "config.h"
#include "common.h"
#include "utils.h"


#define DEFAULT_NRPE_COMMAND	"_NRPE_CHECK"  /* check version of NRPE daemon */

int server_port=DEFAULT_SERVER_PORT;
char *server_name=NULL;
char *command_name=NULL;
int socket_timeout=DEFAULT_SOCKET_TIMEOUT;
int timeout_return_code=STATE_CRITICAL;
int sd;

char query[MAX_INPUT_BUFFER]="";
char respond[MAX_INPUT_BUFFER + 1];

int show_help=FALSE;
int show_license=FALSE;
int show_version=FALSE;

#ifdef HAVE_SSL
SSL_METHOD *meth;
SSL_CTX *ctx;
SSL *ssl;
int use_ssl=TRUE;
#else
int use_ssl=FALSE;
#endif


int process_arguments(int,char **);
void alarm_handler(int);
int graceful_close(int,int);



/* op5 addition */
#define STATE_OOB 4
#define NRPE_MAX_VERSION 2
void print_debug(char *);
struct packet_struct_v1 {
	int packet_type;
	int packet_version;
	int result_code;
	int buffer_length;
	char buffer[MAX_PACKETBUFFER_LENGTH];
};
struct packet_struct_v2 {
	int16_t packet_version;
	int16_t packet_type;
	u_int32_t crc32_value;
	int16_t result_code;
	char buffer[MAX_PACKETBUFFER_LENGTH];
};
typedef struct packet_struct_v1 packet_v1;
typedef struct packet_struct_v2 packet_v2;
char *msg = NULL;
int debug = 0;
int legacy_arg_numbering=0;
int force_ssl = 0;

int main(int argc, char **argv){
        u_int32_t packet_crc32;
        u_int32_t calculated_crc32;
	int16_t result;
	int rc;
	packet send_packet;
	packet receive_packet;
	int bytes_to_send;
	int bytes_to_recv;

	result=process_arguments(argc,argv);

        if(result!=OK || show_help==TRUE || show_license==TRUE || show_version==TRUE){

		if(result!=OK)
			printf("Incorrect command line arguments supplied\n");
                printf("\n");
		printf("NRPE Plugin for Nagios\n");
		printf("Copyright (c) 1999-2008 Ethan Galstad (nagios@nagios.org)\n");
		printf("Version: %s\n",PROGRAM_VERSION);
		printf("Last Modified: %s\n",MODIFICATION_DATE);
		printf("License: GPL v2 with exemptions (-l for more info)\n");
#ifdef HAVE_SSL
		printf("SSL/TLS Available: Anonymous DH Mode, OpenSSL 0.9.6 or higher required\n");
#endif
		printf("\n");
	        }

	if(result!=OK || show_help==TRUE){

		printf("Usage: check_nrpe -H <host> [-n] [-u] [-s] [-o] [-p <port>] [-t <timeout>] [-c <command>] [-a <arglist...>]\n");
		printf("\n");
		printf("Options:\n");
		printf(" -n         = Do no use SSL\n");
		printf(" -s         = Force SSL (and nrpe v2)\n");
		printf(" -o         = Obsolete argument handeling (for compability with old, incorrect, check_nrpe)\n");
		printf(" -u         = Make socket timeouts return an UNKNOWN state instead of CRITICAL\n");
		printf(" <host>     = The address of the host running the NRPE daemon\n");
		printf(" [port]     = The port on which the daemon is running (default=%d)\n",DEFAULT_SERVER_PORT);
		printf(" [timeout]  = Number of seconds before connection times out (default=%d)\n",DEFAULT_SOCKET_TIMEOUT);
		printf(" [command]  = The name of the command that the remote daemon should run\n");
		printf(" [arglist]  = Optional arguments that should be passed to the command.  Multiple\n");
		printf("              arguments should be separated by a space.  If provided, this must be\n");
		printf("              the last option supplied on the command line.\n");
		printf("\n");
		printf("Note:\n");
		printf("This plugin requires that you have the NRPE daemon running on the remote host.\n");
		printf("You must also have configured the daemon to associate a specific plugin command\n");
		printf("with the [command] option you are specifying here.  Upon receipt of the\n");
		printf("[command] argument, the NRPE daemon will run the appropriate plugin command and\n");
		printf("send the plugin output and return code back to *this* plugin.  This allows you\n");
		printf("to execute plugins on remote hosts and 'fake' the results to make Nagios think\n");
		printf("the plugin is being run locally.\n");
		printf("\n");
	        }

	if(show_license==TRUE)
		display_license();

        if(result!=OK || show_help==TRUE || show_license==TRUE || show_version==TRUE)
		exit(STATE_UNKNOWN);


        /* generate the CRC 32 table */
        generate_crc32_table();

#ifdef HAVE_SSL
	/* initialize SSL */
	if(use_ssl==TRUE){
		SSL_library_init();
		SSLeay_add_ssl_algorithms();
		meth=SSLv23_client_method();
		SSL_load_error_strings();
		if((ctx=SSL_CTX_new(meth))==NULL){
			printf("CHECK_NRPE: Error - could not create SSL context.\n");
			exit(STATE_CRITICAL);
		        }

		/* ADDED 01/19/2004 */
		/* use only TLSv1 protocol */
		SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
                }
#endif

	/* initialize alarm signal handling */
	signal(SIGALRM,alarm_handler);

	/* set socket timeout */
	alarm(socket_timeout);

	/* op5 addition                                      */
	/* Split below to connect, send and receive methods  */
	/* This to be able to test differnent nrpe versions  */
	/* and with/without ssl all in one go                */
	int i;
#ifdef HAVE_SSL
	for(use_ssl=force_ssl; use_ssl<2; use_ssl++)
#endif
	for(i=1; i<=NRPE_MAX_VERSION; i++) {
		int res;
		res = nrpe_connect(i, server_name, server_port);
		if(res == -2) {
			print_debug("connect returned -2");
			close(sd);
			continue;
		}

		print_debug("about to send query");
		res = nrpe_send_query(i, (char *)query);
		if(res == -2) {
			close(sd);
			continue;
		}
		
		print_debug("about to reseive response");
		res = nrpe_recv_response(i);
		if(res < STATE_OOB) {
			puts(msg);
			exit(res);
		}

	}

	/* All done... */
	alarm(0);

	if(!msg || !msg[0])
		msg = "Unknown error. Are we allowed to talk to the host?";
	/* this is probably an NRPE error so we add it to the output */
	printf("CHECK_NRPE: %s\n", msg);
	return(STATE_UNKNOWN);
}



/* process command line arguments */
int process_arguments(int argc, char **argv){
	int argindex=0;
	int c=1;
	int i=1;

#ifdef HAVE_GETOPT_LONG
	int option_index=0;
	static struct option long_options[]={
		{"host", required_argument, 0, 'H'},
		{"command", required_argument, 0, 'c'},
		{"args", required_argument, 0, 'a'},
		{"no-ssl", no_argument, 0, 'n'},
		{"ssl", no_argument, 0, 's'},
		{"debug", no_argument, 0, 'd'},
		{"unknown-timeout", no_argument, 0, 'u'},
		{"timeout", required_argument, 0, 't'},
		{"port", required_argument, 0, 'p'},
		{"help", no_argument, 0, 'h'},
		{"license", no_argument, 0, 'l'},
		{"legacy", no_argument, 0, 'o'},
		{0, 0, 0, 0}
                };
#endif

	/* no options were supplied */
	if(argc<2)
		return ERROR;

#define NRPE_OPTCHARS "H:c:a:t:p:dnuhlso"
	while(1){
#ifdef HAVE_GETOPT_LONG
		c=getopt_long(argc,argv,NRPE_OPTCHARS,long_options,&option_index);
#else
		c=getopt(argc,argv,NRPE_OPTCHARS);
#endif
		if(c==-1 || c==EOF)
			break;

		/* process all arguments */
		switch(c){

		case '?':
		case 'h':
			show_help=TRUE;
			break;
		case 'V':
			show_version=TRUE;
			break;
		case 'l':
			show_license=TRUE;
			break;
		case 't':
			socket_timeout=atoi(optarg);
			if(socket_timeout<=0)
				return ERROR;
			break;
		case 'p':
			server_port=atoi(optarg);
			if(server_port<=0)
				return ERROR;
			break;
		case 'H':
			server_name=strdup(optarg);
			break;
		case 'c':
			command_name=strdup(optarg);
			break;
		case 'a':
			argindex=optind;
			break;
		case 'n':
			use_ssl=FALSE;
			force_ssl = 0;
			break;
		case 'u':
			timeout_return_code=STATE_UNKNOWN;
			break;
		case 'd':
			debug++;
			break;
		case 's':
			force_ssl = 1;
			break;
		case 'o':
			legacy_arg_numbering = 1;
			break;
		default:
			return ERROR;
			break;
		        }
	        }

	/* determine (base) command query */
	snprintf(query,sizeof(query),"%s",(command_name==NULL)?DEFAULT_NRPE_COMMAND:command_name);
	query[sizeof(query)-1]='\x0';

	/* If legacy mode (-o) we do not just send the arguments, include part of the */
	/* entire commandline (this to be compatible with old (incorrect) check_nrpe  */
	if(legacy_arg_numbering)
		argindex = 3;
	/* get the command args */
	if(argindex>0){

		for(c=argindex-1;c<argc;c++){

			i=sizeof(query)-strlen(query)-2;
			if(i<=0)
				break;

			strcat(query,"!");
			strncat(query,argv[c],i);
			query[sizeof(query)-1]='\x0';
		        }
	        }

	/* make sure required args were supplied */
	if(server_name==NULL && show_help==FALSE && show_version==FALSE  && show_license==FALSE)
		return ERROR;


	return OK;
        }



void alarm_handler(int sig){

	printf("CHECK_NRPE: Socket timeout after %d seconds.\n",socket_timeout);

	exit(timeout_return_code);
        }


/* submitted by Mark Plaksin 08/31/2006 */
int graceful_close(int sd, int timeout){
        fd_set in;
        struct timeval tv;
        char buf[1000];

#ifdef HAVE_SSL
	if(use_ssl==TRUE){
		SSL_shutdown(ssl);
		SSL_free(ssl);
		SSL_CTX_free(ctx);
		}
#endif

	/* send FIN packet */
        shutdown(sd,SHUT_WR);  
        for(;;){

                FD_ZERO(&in);
                FD_SET(sd,&in);
                tv.tv_sec=timeout/1000;
                tv.tv_usec=(timeout % 1000)*1000;

		/* timeout or error */
                if(1!=select(sd+1,&in,NULL,NULL,&tv))
			break;   

		/* no more data (FIN or RST) */
                if(0>=recv(sd,buf,sizeof(buf),0))
			break;
		}

#ifdef HAVE_CLOSESOCKET
        closesocket(sd);
#else
	close(sd);
#endif

	return OK;
}



/*  Connect to the host */
int nrpe_connect(int version, char *hostname, int port)
{
	int result, rc;
	print_debug("Connecting");

	if(version == 1 && use_ssl) {
		/* ssl wasn't supported until version 2, so simply return -2 and
		 * let the main control loop continue to next version */
		print_debug("v1 do not have ssl");

		return(-2);
	}

	/* try to connect to the host at the given port number */
	result = my_tcp_connect(hostname, port, &sd);

#ifdef HAVE_SSL
	/* do SSL handshake, but fall back to non-ssl if we fail */
	if(result == STATE_OK && use_ssl) {
		if((ssl = SSL_new(ctx)) != NULL) {
			SSL_CTX_set_cipher_list(ctx, "ADH");
			SSL_set_fd(ssl, sd);
			if((rc = SSL_connect(ssl)) != 1) {
				close(sd);
				msg = "Can not connect to a host over SSL.";
				return (-2);
			}
		}
		else {
			printf("CHECK_NRPE: Internal Error - Could not create SSL connection structure.\n");
			exit(STATE_CRITICAL);
		}

		/* bail if we had errors */
		if(result != STATE_OK) {
			SSL_CTX_free(ctx);
		}
	}
	#endif
	if(result != STATE_OK) {
		printf("Unknown connection problem\n");
		exit(STATE_CRITICAL);
	}

	return(sd);
}

/* Send the actual query */
int nrpe_send_query(int version, char *query)
{
	packet_v1 pkt_v1;
	packet_v2 pkt_v2;
	int calculated_crc32;
	int bytes_to_send;
	int pkt_size = 0;
	void *pkt = NULL;
	int rc = 0;

	switch(version) {
	case 1:
		pkt_size = sizeof(packet_v1);
		pkt = &pkt_v1;
		break;
	case 2:
		pkt_size = sizeof(packet_v2);
		pkt = &pkt_v2;
		break;
	default:
		return (-1);
		break;
		}

	/* pkt now points to the correct packet struct type, and
	 * pkt_size is set appropriately, so we can set the required values */
	/* clear the packet buffer */
	memset(pkt, 0, pkt_size);

	/* packet version 1 */
	if(version == 1) {
//		print_debug("Creating v1 type packet");
		memset(&pkt_v1, 0, sizeof(pkt_v1));
		pkt_v1.packet_type = htonl(QUERY_PACKET);
		pkt_v1.packet_version = htonl(NRPE_PACKET_VERSION_1);
		pkt_v1.buffer_length = htonl(strlen(query));
		strcpy(&pkt_v1.buffer[0], query);
		pkt = &pkt_v1;
		pkt_size = sizeof(pkt_v1);
	}

	/* packet version 2 */
	/* fill the packet with semi-random data */
	randomize_buffer((char *)&pkt_v2, sizeof(pkt_v2));
	pkt_v2.packet_version = (int16_t)htons(NRPE_PACKET_VERSION_2);
	pkt_v2.packet_type = (int16_t)htons(QUERY_PACKET);
	strncpy(&pkt_v2.buffer[0], query, MAX_PACKETBUFFER_LENGTH);
	pkt_v2.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';
	/* calculate the crc 32 value of the packet */
	pkt_v2.crc32_value = (u_int32_t) 0L;
	calculated_crc32 = calculate_crc32((char *)pkt, pkt_size);
	pkt_v2.crc32_value = (u_int32_t) htonl(calculated_crc32);


	/***** ENCRYPT REQUEST *****/

	bytes_to_send = pkt_size;

	/* send the packet */
# ifdef HAVE_SSL
	print_debug("Encrypting packet");
	if(use_ssl) {
		if(version == 2) {
			rc = SSL_write(ssl, (char *)&pkt_v2, pkt_size);
		}
		else {
			/* fallthrough. should never happen, if compatibility code
			 * is added properly */
			rc = SSL_write(ssl, pkt, pkt_size);
		}
		if(rc < 0) rc = -1;
	}
	else
# endif
		switch(version) {
		case 1:
			print_debug("Sending v1 type packet");
			rc = sendall(sd, (char *)&pkt_v1, &bytes_to_send);
			break;
		case 2:
			print_debug("Sending v2 type packet");
			rc = sendall(sd, (char *)&pkt_v2, &bytes_to_send);
			break;
		default:
			print_debug("Sending default type packet");
			rc = sendall(sd, (char *)&pkt, &bytes_to_send);
			break;
		}

	if(rc == -1) {
//		msg = "Error sending query to host.\n";
		printf("Error sending query to host.\n");
		return STATE_UNKNOWN;
	}
	return(rc);
}


/* Wait for and receive response */
// int nrpe_recv_response(unsigned short version)
int nrpe_recv_response(int version)
{
	int pkt_size = 0;
	int pkt_type = 0;
	void *pkt = NULL;
	packet_v1 pkt_v1;
	packet_v2 pkt_v2;
	int rc;
	int bytes_to_receive;
	int packet_crc32;
	int calculated_crc32;
	int result = 0;
	char *pbuffer=&respond[0];
	char *buf;
	int len = 0;
	int buflen = 0;

	switch(version) {
	case 1:
		use_ssl = 0;
		pkt = &pkt_v1;
		pkt_size = sizeof(pkt_v1);
		break;
	case 2:
		pkt = &pkt_v2;
		pkt_size = sizeof(pkt_v2);
		break;
	default:
		return (-1);
	}

	/* Altinity patch: Allow multiple packets to be received */
	/* Indentation not corrected to allow simpler patching */
	/* START MULTI_PACKET LOOP */
	do {

	bytes_to_receive = pkt_size;

# ifdef HAVE_SSL
	if(use_ssl) {
		rc = SSL_read(ssl, pkt, pkt_size);
//		printf("SSL_read(): %d, SSL_get_error(): %d\n",
//			   rc, SSL_get_error(ssl, rc));
	}
	else
# endif
		rc = recvall(sd, (char *)pkt, (int *)&bytes_to_receive, socket_timeout);

	/* recv() error. This is OK. Just try next version */
	if(rc == -2) {
		msg = "";
		print_debug("Receive failed, continuing");
		graceful_close(sd,1000);
		return STATE_OOB;
	}
	else if(rc == -1) {
		msg = "Error receiving data from daemon.";
		print_debug("Receive failed, continuing");
		graceful_close(sd,1000);
		return STATE_OOB;
	}
	else if(rc == 0) {
		/* server disconnected */
		print_debug("Receive failed, continuing");
		msg = "Received 0 bytes from daemon. Check remote server logs for error messages.";
		graceful_close(sd,1000);
		return STATE_OOB;
	}

	/* receive underflow */
	else if(bytes_to_receive < pkt_size) {
		msg = "Receive underflow. Failed to receive entire buffer.";
		print_debug("Receive failed, continuing");
		graceful_close(sd,1000);
		return STATE_OOB;
	}

	/* packet validation */
	if(version == 1) {
		result = ntohl(pkt_v1.result_code);
		if(!strlen(pkt_v1.buffer)) {
			msg = "No output from remote host.";
		}
		pkt_type = pkt_v1.packet_type;
		buf = pkt_v1.buffer;
	}
	else if(version == 2) {
		packet_crc32 = ntohl(pkt_v2.crc32_value);
		pkt_v2.crc32_value = 0L;
		calculated_crc32 = calculate_crc32((char *)pkt, pkt_size);
		if(packet_crc32 != calculated_crc32) {
			msg = "Response packet had invalid CRC32.";
//			puts(msg);
			graceful_close(sd,1000);
			return STATE_OOB;
		}
		if(ntohs(pkt_v2.packet_version) != NRPE_PACKET_VERSION_2) {
//			puts(msg);
			msg = "Invalid packet version received from server.";
			graceful_close(sd,1000);
			return STATE_OOB;
		}
		pkt_type = ntohs(pkt_v2.packet_type);
		if(pkt_type != RESPONSE_PACKET && pkt_type != RESPONSE_PACKET_WITH_MORE) {
//			puts(msg);
			msg = "Invalid packet type received from server.";
			graceful_close(sd,1000);
			return STATE_OOB;
		}
		result = (int16_t)ntohs(pkt_v2.result_code);
		pkt_v2.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';
		buf = pkt_v2.buffer;
	}
	buflen = strlen(buf);
	if (buflen > MAX_INPUT_BUFFER - len) {
		buflen = MAX_INPUT_BUFFER - len;
	}
	memcpy(pbuffer, buf, buflen);
	pbuffer += buflen;
	len += buflen;

	} while (pkt_type==RESPONSE_PACKET_WITH_MORE && len<MAX_INPUT_BUFFER);
	/* END MULTI_PACKET LOOP */

	/* Finish output with newline */
	if (len < MAX_INPUT_BUFFER && respond[len-1] != '\n') {
		respond[len] = '\n';
		len++;
	}
	respond[len] = '\x0';
	msg = respond;

	/* close the connection */
	graceful_close(sd,1000);

	return result;
}

/* Helper, print debug info */
void print_debug(char* str)
{
	if(debug != 0) {
		printf("%s\n", str);
	}
}
